/* 
 * Licensed to the Apache Software Foundation (ASF) under one or more
 * contributor license agreements.  See the NOTICE file distributed with
 * this work for additional information regarding copyright ownership.
 * The ASF licenses this file to You under the Apache License, Version 2.0
 * (the "License"); you may not use this file except in compliance with
 * the License.  You may obtain a copy of the License at
 * 
 *     http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package org.apache.harmony.sql.internal.rowset;

import java.sql.SQLException;
import java.util.BitSet;

public class CachedRow implements Cloneable {
    private Object[] columnData;

    private Object[] originalColumnData;

    private BitSet mask;

    private boolean isDelete;

    private boolean isInsert;

    private boolean isUpdate;

    private boolean nonUpdateable = false;

    public CachedRow(Object[] columnData) {
        this.columnData = columnData.clone();
        originalColumnData = columnData.clone();
        mask = new BitSet(columnData.length);
    }

    public boolean getUpdateMask(int i) {
        return mask.get(i);
    }

    public void setUpdateMask(int i) {
        mask.set(i);
    }

    public void setUnavailable() {
        // FIXME: What is this method used for?
        setDelete();
        setInsert();
        mask.clear();
        mask.flip(0, columnData.length);
    }

    public void setNonUpdateable() {
        // setDelete();
        // setInsert();
        // mask.clear();
        // mask.flip(0,columnData.length);
        nonUpdateable = true;
    }

    public boolean getNonUpdateable() {
        return nonUpdateable;
    }

    public void setDelete() {
        isDelete = true;
        isUpdate = false;
        isInsert = false;
    }

    public void undoDelete() {
        isDelete = false;
    }

    public boolean isDelete() {
        return isDelete;
    }

    public void setInsert() {
        isInsert = true;
        isUpdate = false;
        isDelete = false;
    }

    public boolean isInsert() {
        return isInsert;
    }

    public void setUpdate() {
        isUpdate = true;
        isInsert = false;
        isDelete = false;
    }

    public void undoUpdate() {
        isUpdate = false;
        mask.flip(0, columnData.length);
        columnData = originalColumnData.clone();
    }

    public boolean isUpdate() {
        return isUpdate;
    }

    public void updateObject(int columnIndex, Object x) throws SQLException {
        if (nonUpdateable) {
            // TODO load message from resource file
            throw new SQLException("Not Updateable of the CurrentRow");
        }

        columnData[columnIndex - 1] = x;
        setUpdateMask(columnIndex - 1);
    }

    public CachedRow getOriginal() {
        return new CachedRow(originalColumnData);
    }

    public Object getObject(int columnIndex) {
        return columnData[columnIndex - 1];
    }

    // deep clone
    public CachedRow createClone() throws CloneNotSupportedException {
        CachedRow cr = (CachedRow) super.clone();

        Object[] cd = new Object[columnData.length];
        for (int i = 0; i < columnData.length; i++) {
            cd[i] = columnData[i];
        }
        cr.columnData = cd;
        cr.isInsert = isInsert;
        cr.isDelete = isDelete;
        cr.isUpdate = isUpdate;
        cr.mask = (BitSet) mask.clone();
        cr.nonUpdateable = nonUpdateable;
        // cr.originalColumnData
        return cr;
    }
}
